local mod = {}

local SHOT_SOUND = sound.load("asteroid_shot.wav")
local EXPLOSION_SOUND = sound.load("asteroid_explosion.wav")
local THRUST_SOUND = sound.load("asteroid_thrust.wav")
sound.loop(THRUST_SOUND)

local BACKGROUND_TEXTURE = image.load("asteroid.png")
local ASTEROID_BIG_TEXTURE = image.load("asteroid_big.png")
image.center(ASTEROID_BIG_TEXTURE)
local ASTEROID_SMALL_TEXTURE = image.load("asteroid_small.png")
image.center(ASTEROID_SMALL_TEXTURE)
local ASTEROID_BULLET_TEXTURE = image.load("asteroid_bullet.png")
image.center(ASTEROID_BULLET_TEXTURE)
local ASTEROID_SHIP_TEXTURE = image.load("asteroid_ship.png")
image.center(ASTEROID_SHIP_TEXTURE)

local ROTATIONAL_SPEED = math.pi / 30

local SHIP_RADIUS = 12
local SHIP_HITBOX_RADIUS = 8 -- Give a small margin for hits
local SHIP_COLOR = color.new(0, 255, 0)
local SHIP_MAX_SPEED = 5
local SHIP_DECELERATION = 0.1
local SHIP_ACCELERATION = 0.18

local ASTEROID_MIN_DISTANCE = 50
local ASTEROID_COLOR = color.new(255, 255, 255)
local ASTEROID_SPEED = 0.6
local ASTEROID_BIG_RADIUS = 24
local ASTEROID_SMALL_RADIUS = 12

local ASTEROID_POINTS_BIG = 20
local ASTEROID_POINTS_SMALL = 10
local ASTEROID_POINTS_GAME = 70

local BULLET_COLOR = color.new(255, 255, 0)
local BULLET_RADIUS = 3
local BULLET_BASE_SPEED = 6

local PLAYFIELD_WIDTH = 260
local PLAYFIELD_HEIGHT = 260

local DRAW_X_OFFSET = 480 / 2 - PLAYFIELD_WIDTH / 2
local DRAW_Y_OFFSET = 272 / 2 - PLAYFIELD_HEIGHT / 2

local function circles_collide(a_x, a_y, a_rad, b_x, b_y, b_rad)
    return (a_x - b_x) * (a_x - b_x) + (a_y - b_y) * (a_y - b_y) <
               (a_rad + b_rad) * (a_rad + b_rad)
end

local function set_asteroids(STATE)
    math.randomseed(0)
    STATE.asteroid_asteroids = {}
    for i = 1, STATE.asteroid_game_level do
        local x = math.random(0, PLAYFIELD_WIDTH - 1)
        local y = math.random(0, PLAYFIELD_HEIGHT - 1)
        while circles_collide(x, y, ASTEROID_BIG_RADIUS, STATE.asteroid_ship_x,
                              STATE.asteroid_ship_y, ASTEROID_MIN_DISTANCE) do
            x = math.random(0, PLAYFIELD_WIDTH - 1)
            y = math.random(0, PLAYFIELD_HEIGHT - 1)
        end
        local angle = math.random() * math.pi * 2 -- TODO prevent going into the ship if too close?
        STATE.asteroid_asteroids[i] = {
            x = x,
            y = y,
            angle = angle,
            radius = ASTEROID_BIG_RADIUS
        }
    end
end

function mod.reset(STATE)
    STATE.asteroid_ship_x = PLAYFIELD_WIDTH / 2
    STATE.asteroid_ship_y = PLAYFIELD_HEIGHT / 2
    STATE.asteroid_ship_angle = math.pi / 2
    STATE.asteroid_ship_speed = 0
    STATE.asteroid_game_level = 1
    STATE.asteroid_bullets = {}
    set_asteroids(STATE)
end

function mod.render(STATE)
    image.blit(BACKGROUND_TEXTURE, 0, 0)

    screen.clip(DRAW_X_OFFSET, DRAW_Y_OFFSET, PLAYFIELD_WIDTH, PLAYFIELD_HEIGHT)
    for _, asteroid in ipairs(STATE.asteroid_asteroids) do
        if asteroid.radius == ASTEROID_SMALL_RADIUS then
            image.blit(ASTEROID_SMALL_TEXTURE, DRAW_X_OFFSET + asteroid.x,
                       DRAW_Y_OFFSET + asteroid.y)
        else
            image.blit(ASTEROID_BIG_TEXTURE, DRAW_X_OFFSET + asteroid.x,
                       DRAW_Y_OFFSET + asteroid.y)
        end
    end
    for _, bullet in ipairs(STATE.asteroid_bullets) do
        image.blit(ASTEROID_BULLET_TEXTURE, bullet.x + DRAW_X_OFFSET,
                   bullet.y + DRAW_Y_OFFSET)
    end
    image.rotate(ASTEROID_SHIP_TEXTURE,
                 STATE.asteroid_ship_angle / math.pi * 180)
    image.blit(ASTEROID_SHIP_TEXTURE, STATE.asteroid_ship_x + DRAW_X_OFFSET,
               STATE.asteroid_ship_y + DRAW_Y_OFFSET)
    screen.clip()
end

local function update_coords(x, y, radius, angle, speed)
    x = x + math.cos(angle) * speed
    y = y + math.sin(angle) * speed
    local looped = false

    if x < -radius then
        x = x + PLAYFIELD_WIDTH + radius * 2
        looped = true
    end
    if x > PLAYFIELD_WIDTH + radius then
        x = x - PLAYFIELD_WIDTH - radius * 2
        looped = true
    end

    if y < -radius then
        y = y + PLAYFIELD_HEIGHT + radius * 2
        looped = true
    end
    if y > PLAYFIELD_HEIGHT + radius then
        y = y - PLAYFIELD_HEIGHT - radius * 2
        looped = true
    end

    return x, y, looped
end

function mod.update(STATE)
    -- Inputs
    if (buttons.held.left or buttons.analogx < -20) then
        STATE.asteroid_ship_angle = STATE.asteroid_ship_angle - ROTATIONAL_SPEED
    end
    if (buttons.held.right or buttons.analogx > 20) then
        STATE.asteroid_ship_angle = STATE.asteroid_ship_angle + ROTATIONAL_SPEED
    end
    if buttons.held.cross then
        STATE.asteroid_ship_speed = math.min(
                                        STATE.asteroid_ship_speed +
                                            SHIP_ACCELERATION, SHIP_MAX_SPEED)
        if not sound.playing(THRUST_SOUND) then sound.play(THRUST_SOUND) end
    else
        STATE.asteroid_ship_speed = math.max(
                                        STATE.asteroid_ship_speed -
                                            SHIP_DECELERATION, 0)
        sound.stop(THRUST_SOUND)
    end
    if buttons.circle then
        local ship_tip_x = math.cos(STATE.asteroid_ship_angle) * SHIP_RADIUS +
                               STATE.asteroid_ship_x
        local ship_tip_y = math.sin(STATE.asteroid_ship_angle) * SHIP_RADIUS +
                               STATE.asteroid_ship_y
        STATE.asteroid_bullets[#STATE.asteroid_bullets + 1] = {
            x = ship_tip_x,
            y = ship_tip_y,
            angle = STATE.asteroid_ship_angle,
            speed = STATE.asteroid_ship_speed + BULLET_BASE_SPEED
        }
        sound.play(SHOT_SOUND)
    end

    -- Coordinates updates
    STATE.asteroid_ship_x, STATE.asteroid_ship_y, _ = update_coords(
                                                          STATE.asteroid_ship_x,
                                                          STATE.asteroid_ship_y,
                                                          SHIP_RADIUS,
                                                          STATE.asteroid_ship_angle,
                                                          STATE.asteroid_ship_speed)
    for _, asteroid in ipairs(STATE.asteroid_asteroids) do
        asteroid.x, asteroid.y, _ = update_coords(asteroid.x, asteroid.y,
                                                  asteroid.radius,
                                                  asteroid.angle, ASTEROID_SPEED)
    end
    local bullets_to_remove = {}
    for i, bullet in ipairs(STATE.asteroid_bullets) do
        new_x, new_y, looped = update_coords(bullet.x, bullet.y, BULLET_RADIUS,
                                             bullet.angle, bullet.speed)
        if looped then
            bullets_to_remove[#bullets_to_remove + 1] = i
        else
            bullet.x = new_x
            bullet.y = new_y
        end
    end
    for i = #bullets_to_remove, 1, -1 do
        table.remove(STATE.asteroid_bullets, bullets_to_remove[i])
    end

    -- Collisions
    local asteroids_to_destroy = {}
    local bullets_to_remove = {}
    for i, asteroid in ipairs(STATE.asteroid_asteroids) do
        -- ship hit asteroid, game lost
        if circles_collide(asteroid.x, asteroid.y, asteroid.radius,
                           STATE.asteroid_ship_x, STATE.asteroid_ship_y,
                           SHIP_HITBOX_RADIUS) then return true end
        -- bullets
        for j, bullet in ipairs(STATE.asteroid_bullets) do
            if circles_collide(bullet.x, bullet.y, BULLET_RADIUS, asteroid.x,
                               asteroid.y, asteroid.radius) then
                asteroids_to_destroy[#asteroids_to_destroy + 1] = i
                bullets_to_remove[#bullets_to_remove + 1] = j
                sound.play(EXPLOSION_SOUND)
                break
            end
        end
    end
    for i = #bullets_to_remove, 1, -1 do
        table.remove(STATE.asteroid_bullets, bullets_to_remove[i])
    end
    local new_asteroids = {}
    for i = #asteroids_to_destroy, 1, -1 do
        asteroid = table.remove(STATE.asteroid_asteroids,
                                asteroids_to_destroy[i])
        if asteroid.radius == ASTEROID_BIG_RADIUS then
            STATE.game_score = STATE.game_score + ASTEROID_POINTS_BIG
            new_asteroids[#new_asteroids + 1] = {
                x = asteroid.x,
                y = asteroid.y,
                radius = ASTEROID_SMALL_RADIUS,
                angle = math.random() * math.pi * 2
            }
            new_asteroids[#new_asteroids + 1] = {
                x = asteroid.x,
                y = asteroid.y,
                radius = ASTEROID_SMALL_RADIUS,
                angle = math.random() * math.pi * 2
            }
        else
            STATE.game_score = STATE.game_score + ASTEROID_POINTS_SMALL
        end
    end
    for i = 1, #new_asteroids do
        STATE.asteroid_asteroids[#STATE.asteroid_asteroids + 1] =
            new_asteroids[i]
    end

    -- Check if game won
    if #STATE.asteroid_asteroids == 0 then
        STATE.asteroid_ship_x = PLAYFIELD_WIDTH / 2
        STATE.asteroid_ship_y = PLAYFIELD_HEIGHT / 2
        STATE.asteroid_ship_angle = math.pi / 2
        STATE.asteroid_ship_speed = 0
        STATE.asteroid_game_level = STATE.asteroid_game_level + 1
        STATE.asteroid_bullets = {}
        STATE.game_score = STATE.game_score + ASTEROID_POINTS_GAME
        set_asteroids(STATE)
    end
end

function mod.stop_sounds()
    sound.stop(SHOT_SOUND)
    sound.stop(EXPLOSION_SOUND)
    sound.stop(THRUST_SOUND)
end

return mod
